
import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;

import java.util.ArrayList;
import java.util.List;

import org.apache.poi.poifs.filesystem.POIFSFileSystem;
import org.apache.poi.poifs.filesystem.DocumentEntry;
import org.apache.poi.util.LittleEndian;

import org.apache.poi.hpsf.PropertySet;
import org.apache.poi.hpsf.SummaryInformation;
import org.apache.poi.hpsf.DocumentSummaryInformation;


/**
 * <B>DOCHandler</B> is content handler for Microsoft Word 8.0/97-2000 documents
 * (application/msword content type). Currently, Microsoft Word 6.0/95 documents
 * are only partially supported - text is propperly extracted but there is no
 * ANSI to Unicode conversion (character code pages are ignored). Therefore,
 * Word 6.0/95 documents will be properlly processed only if they are written
 * using characters from the first code page (i.e. written in english, german...)
 * DOES NOT WORK WITH WORD XP!
 * 
 * @author  Miroslav Obradovic
 * @e-mail  micky@eunet.yu
 * @date    21/11/2002
 * @version 1.0
 */
public final class DOCHandler {
    
    // Content data.
    private String title;             
    private String description;       
    private String keywords;          
    private String categories;        
    private long published;           
    private String author;            
    private StringBuffer contents;    

    // POI file system (DOC file in OLE 2.0 format).
    private POIFSFileSystem filesystem;
    
    // Byte buffer containing the main stream of a DOC file.
    private byte[] _mainStream;
    
    // FIB (File Information Block) data.
    private int _nFib = 0;            // FIB version (see method isWord97Format()).
    private boolean _fComplex;        // Is file in complex (fast-saved) format?
    private int _fcMin;               // The beginning of text in non-complex document.
    private int _fcMac;               // The end of text in non-complex document.
    private int _fcClx;               // The beginning of complex data in complex document.
    private boolean _fWhichTblStm;    // Indicates which table stream is valid.
    
    
    /**
     * Initialize variables.
     */
    public DOCHandler() {
        
        contents = new StringBuffer();
        reset();
    }
    
    /**
     * Return document title.
     *
     *     @return    always null
     */
    public String getTitle() {
        
        return this.title;
    }
    
    /**
     * Return document description.
     *
     *     @return    always null
     */
    public String getDescription() {
        
        return this.description;
    }
    
    /**
     * Return document keywords.
     *
     *     @return    always null
     */
    public String getKeywords() {
        
        return this.keywords;
    }
    
    /**
     * Return document categories.
     *
     *     @return    always null
     */
    public String getCategories() {
        
        return this.categories;
    }
    
    /**
     * Return document published date.
     *
     *     @return    always -1
     */
    public long getPublished() {
        
        return this.published;
    }
    
    /**
     * Return document author.
     *
     *     @return    always null
     */
    public String getAuthor() {
        
        return this.author;
    }
    
    /**
     * Return document contents.
     *
     *     @return    document contents
     */
    public String getContents() {
        
        return this.contents.toString();
    }
    
    /**
     * Replace characters with codes less than 32 using blanks.
     */
    private void removeControlCharacters() {
        
        if (contents == null) {
            
            return;
        }
        
        StringBuffer newContents = new StringBuffer();
        for (int i = 0; i < contents.length(); i++) {
            
            char ch = contents.charAt(i);
            newContents.append(ch < 32 ? ' ' : ch);
        }
        contents = newContents;
    }
    
    /**
     * Read document properties using POI HPSF (Horrible Property Set Format) package.
     */
    private void readProperties() {
        
        // Get the SIS (Summary Information Stream) and create the Summary Information.
        SummaryInformation summaryInformation = null;
        byte[] SIS = null;
        try {
            
            // Allocate enough buffer space and get the stream.
            // Method getSize() returns the count of bytes.
            DocumentEntry documentEntry = 
                    (DocumentEntry) filesystem.getRoot().getEntry("\005SummaryInformation");
            SIS = new byte[documentEntry.getSize()];
            filesystem.createDocumentInputStream("\005SummaryInformation").read(SIS);
            summaryInformation = new SummaryInformation(new PropertySet(SIS));
            
        } catch (java.lang.Exception e) {
            
            summaryInformation = null;
        }
        
        // Get needed data if Summary Information was successfully created.
        if (summaryInformation != null) {
            
            // Get available properties.
            title = summaryInformation.getTitle();
            description = summaryInformation.getSubject();
            keywords = summaryInformation.getKeywords();
            author = summaryInformation.getAuthor();
            
            // If there's no subject, use comments as description if any.
            if (description == null) {
                
                description = summaryInformation.getComments();
            }
            
            // Get published.
            java.util.Date date;
            date = summaryInformation.getLastSaveDateTime();  // get Modified
            if (date == null) {
                
                date = summaryInformation.getCreateDateTime();  // get Created
            }
            
            if (date != null) {
                
                published = date.getTime();
                
            } else {
                
                published = -1;
            }
        }

        // MyDocumentSummaryInformation class is a manual work around
        // the POI HPSF DocumentSummaryInformation.getCategory() bug.
        // See MyDocumentSummaryInformation for details.
        
        // HERE'S MyDocumentSummaryInformation INSTEAD OF DocumentSummaryInformation.
        MyDocumentSummaryInformation documentSummaryInformation = null;
        
        // Get the DSIS (Document Summary Information Stream) and create 
        // the Document Summary Information.
        byte[] DSIS = null;
        try {
            
            // Allocate enough buffer space and get the stream.
            // Method getSize() returns the count of bytes.
            DocumentEntry documentEntry = 
                    (DocumentEntry) filesystem.getRoot().getEntry("\005DocumentSummaryInformation");
            DSIS = new byte[documentEntry.getSize()];
            filesystem.createDocumentInputStream("\005DocumentSummaryInformation").read(DSIS);
            
            // HERE'S MyDocumentSummaryInformation INSTEAD OF DocumentSummaryInformation.
            documentSummaryInformation = new MyDocumentSummaryInformation(new PropertySet(DSIS));
            
        } catch (java.lang.Exception e) {
            
            documentSummaryInformation = null;
        }
        
        // Get needed data if Document Summary Information was successfully created.
        if (documentSummaryInformation != null) {
            
            categories = documentSummaryInformation.getCategory();
        }
    }
    
    /**
     * Goes through the piece table and parses out the info regarding the text
     * blocks.
     *
     *     @param     tableStream      buffer containing the main table stream
     *     @param     complexOffset    beginning of the complex data
     */
    private void extractTextPieces(byte[] tableStream, int complexOffset) {
        
        // Each block in the CLX (CompLeX) part of the table stream is prefaced by
        // a CLXT (CLX Type), which is a 1-byte code, either 1 (clxtGrpprl meaning
        // the block contains a GRPPRL [GRouP of PRoperty modifiers stored in a List,
        // i.e. group of SPRMs]) or 2 (clxtPlcfpcd, meaning this is the PLCFPCD, 
        // i.e. the piece table itself).
        
        int pos = complexOffset;    // current position within the table stream
        
        // Skip through the PRMs (PRoperty Modifiers) before the piece table is reached.
        // Each clxtGrpprl (1) is followed by a 2-byte CB (Count of Bytes) of the GRPPRL.
        // A full saved file will have no clxtGrpprl's.
        while (tableStream[pos] == 1) {    // (1 == clxtGrpprl)
            
            int skip = LittleEndian.getShort(tableStream, ++pos);    // 16 bits (short)
            pos += 2 + skip;    // skip CB (two bytes) and the whole GRPPRL
        }
        
        // PLCFPCD (piece table) must be present in a complex document after GRPPRLs.
        if (tableStream[pos] != 2) {    // (2 == clxtPlcfpcd)
            
            // Corrupted Word file - do nothing.
            
        } else {
            
            // A clxtPlcfpcd (2) is followed by a 4-byte LCB ([Long] Count of
            // Bytes) of the piece table.
            // CB is a 2-byte value in Word 6.0/95 compatible format!
            int pieceTableSize;
            if (isWord97Format()) {    // Word 8.0/97 compatible format
                
                pieceTableSize = LittleEndian.getInt(tableStream, ++pos);    // 32 bits (int)
                pos += 4;    // skip LCB, now 'pos' indicates the beginning of PLCFPCD
                
            } else {    // Word 6.0/95 compatible format
                
                pieceTableSize = LittleEndian.getShort(tableStream, ++pos);    // 16 bits (short)
                pos += 2;    // skip CB, now 'pos' indicates the beginning of PLCFPCD
            }
            
            // Note: The piece table is a data structure that describes the logical
            // sequence of characters in a Word document and records recent changes
            // to the formatting of a Word document. It is stored in a Word file as
            // a PLCF (PLex of Cps stored in File) named the PLCFPCD (PLCF containing
            // PieCe Descriptors). The piece table relates a logical character number,
            // called a CP (Character Position), to a physical location within a Word
            // file (FC). The array of CPs in the PLCFPCD defines a partitioning of the
            // Word document into disjoint pieces. The second array is an array of PCDs
            // (Piece Descriptors) which is in 1-to-1 correspondence to the array of CPs
            // and records the physical location in the Word file where the corresponding
            // piece begins.
                        
            // Calculate index of the last piece stored in PLCFPCD using:
            //     iMac = (pieceTableSize - 4) / (4 + cbPCD);
            // where cbPCD (count of bytes of PCD) equals 8.
            // In Word 6.0/95 compatible format, count of bytes of PCD equals 10!
            int iMac;
            if (isWord97Format()) {    // Word 8.0/97 compatible format
                
                iMac = (pieceTableSize - 4) / 12;
                
            } else {    // Word 6.0/95 compatible format
                
                iMac = (pieceTableSize - 4) / 14;
            }
            
            // Calculate offset of the first CP and the first PCD in the piece table.
            // The number of CPs (Character Positions) stored in the PLCF will be (iMac + 1).
            // Array of PCDs starts at offset (iMac + 1) * 4 because the size of single CP
            // is 4 bytes.
            int currentCP = pos;
            int currentPCD = pos + (iMac + 1) * 4;
            
            // Parse out the text pieces.
            for (int i = 0; i <= iMac; i++) {
                
                // Get FC field (file offset of beginning of piece) from the current PCD.
                int filePos = LittleEndian.getInt(tableStream, currentPCD + 2);  // 32 bits (int)
                
                // If the second most significant bit of 'filePos' is clear, then this
                // is the unicode character (two bytes). If the second most significant
                // bit is set, the actual address of the [codepage-1252 compressed
                // version of the unicode] character (one byte) is actually at the offset
                // indicated by clearing this bit and dividing by two.
                boolean unicode = false;
                if (isWord97Format()) {    // unicode only in Word 8.0/97 compatible format
                    
                    if ((filePos & 0x40000000) == 0) {
                        
                        unicode = true;
                        
                    } else {
                        
                        filePos &= ~(0x40000000);
                        filePos /= 2;
                    }
                }
                
                // The size of the i-th piece can be determined by subtracting rgcp[i] of
                // the containing PLCFPCD from it's rgcp[i+1]. Note: 'rg' is a prefix used
                // to signify that the data structure being defined is an array (e.g. 'rgcp'
                // is an array of CPs). Size of a single CP is 4 bytes (32 bits [int]).
                int pieceSize = LittleEndian.getInt(tableStream, currentCP + 4) -
                        LittleEndian.getInt(tableStream, currentCP);
                
                // Append the current text piece to contents.
                char ch;
                if (unicode) {
                    
                    for (int j = filePos; j < filePos + pieceSize; j += 2) {
                        
                        ch = (char) LittleEndian.getUShort(_mainStream, j);    // 16 bits (char)
                        contents.append(ch);
                    }
                    
                } else {
                    
                    
                    for (int j = filePos; j < filePos + pieceSize; j += 1) {
                        
                        ch = (char) _mainStream[j];    // 8 bits (char)
                        contents.append(ch);
                    }
                }
                
                // Skip to the next piece. Size of a single CP is 4 bytes.
                // Size of a single PCD is 8 bytes in Word 8.0/97, and 10
                // bytes in Word 6.0/95 compatible format.
                currentCP += 4;
                currentPCD += (isWord97Format() ? 8 : 10);
            }
        }
    }
    
    /**
     * Extract text from a complex document.
     */
    private void processComplexDocument() {
        
        if (isWord97Format()) {    // Word 8.0/97 compatible format
            
            // Get the valid table stream name.
            String tableName = null;
            if (_fWhichTblStm) {
                
                tableName = new String("1Table");
                
            } else {
                
                tableName = new String("0Table");
            }
            
            // Get the document entry containing the table stream.
            byte[] tableStream = null;
            try {
                
                DocumentEntry tableEntry = (DocumentEntry) filesystem.getRoot().getEntry(tableName);
                
                // Load the table stream into a buffer. Method getSize() returns the size in bytes.
                int size = tableEntry.getSize();
                tableStream = new byte[size];
                filesystem.createDocumentInputStream(tableName).read(tableStream);
                
            } catch (java.lang.Exception e) {
                
                tableStream = null;
            }
            
            // Extract text pieces from the document and append them to contents.
            if (tableStream != null) {
                
                extractTextPieces(tableStream, _fcClx);
            }
            
        } else {    // Word 6.0/95 compatible format
            
            extractTextPieces(_mainStream, _fcClx);
        }
    }
    
    /**
     * Extract text from non-complex document. This method has to be improved
     * to better suit Word 6.0/95 documents. When extracting text from Word 6.0/95 
     * compatible documents, characters should be translated to the appropriate
     * code page, i.e. an [ANSI + language + character set] to [Unicode] conversion
     * is needed. Word 8.0/97 compatible documents use Unicode, so this method
     * works properly for them.
     */
    private void processNonComplexDocument() {
        
        // Extract the whole text stream from the document.
        int size = _fcMac - _fcMin;
        byte[] text = new byte[size];
        System.arraycopy(_mainStream, _fcMin, text, 0, size);
        
        // Copy text to contents buffer.
        contents = new StringBuffer(new String(text));
    }
    
    /**
     * Check if used FIB format is Word 97 compatible. This method must NOT be
     * called before a call to readFIB(). <BR>
     * <BR>
     * Known FIB versions: <BR>
     * <UL>
     *  <LI> <B>104</B> Word 6.0/95 compatible format </LI>
     *  <LI> <B>193</B> Word 8.0/97 compatible format </LI>
     * </UL>
     */
    private boolean isWord97Format() {
        
        boolean result = false;
        
        switch (_nFib) {
            
            case 0 :         // method is called before a call to readFIB(), error
                break;
                
            case 104:        // Word 6.0/95 compatible format
                result = false;
                break;
                
            case 193:        // Word 8.0/97 compatible format
                result = true;
                break;
                
            default:         // all other cases, try to guess ;-)))
                result = (_nFib >= 105 ? true : false);
                break;
        }
        
        return result;
    }
    
    /**
     * Extracts the main document stream and needed FIB data from the POI file. <BR>
     * <BR>
     * <B>Note:</B> <I>FIB = File Information Block</I> is the header of a Word file.
     * The FIB starts at the beginning of the file. Gives the beginning offset and
     * lengths of the document's text stream and subsidiary data structures within
     * the file. Also stores other file status information.
     */
    private void readFIB() {
        
        try {
            
            // Allocate enough buffer space and get the document's main stream.
            // Method getSize() returns the document size in bytes.
            DocumentEntry documentEntry = 
                    (DocumentEntry) filesystem.getRoot().getEntry("WordDocument");
            _mainStream = new byte[documentEntry.getSize()];
            filesystem.createDocumentInputStream("WordDocument").read(_mainStream);
            
        } catch (java.lang.Exception e) {
            
            _mainStream = null;
        }
        
        // Get needed information from FIB (File Information Block).
        if (_mainStream != null) {
            
            // Get the FIB version. See method isWord97format() for specific values.
            int _nFib = LittleEndian.getShort(_mainStream, 0x02);      // 16 bits (short)
            
            //System.out.println("FIB version: " + _nFib);
            
            // Get flags.
            int flags = LittleEndian.getShort(_mainStream, 0x0a);      // 16 bits (short)
            
            // If (_fComplex), the document text stream is represented by the text
            // beginning at _fcMin up to (but not including) _fcMac.
            // Otherwise, i.e. if (!_fComplex), file is in complex, fast-saved format
            // and the document is represented by the piece table stored in the table
            // stream within the data beginning at _fcClx.
            _fComplex = (flags & 0x0004) != 0;
            
            // File offset of the first character of text in a non-complex document.
            _fcMin = LittleEndian.getInt(_mainStream, 0x18);           // 32 bits (int)
            
            // File offset of the last character of text in a non-complex document + 1.
            _fcMac = LittleEndian.getInt(_mainStream, 0x1c);           // 32 bits (int)
            
            // Offset in the table stream of beginning of information for complex files
            // which consists of an encoding of all of the PRMs (PRoperty Modifier) quoted
            // by the document followed by the piece table for the document.
            // In Word '97 and later, the _fcClx always indicates the location of the
            // "complex" table information and it is always necessary to determine
            // the location and contents of text and properties. This arises due to
            // unicode and unicode compression.
            if (isWord97Format()) {    // Word 8.0/97 compatible format
                
                _fcClx = LittleEndian.getInt(_mainStream, 0x01a2);         // 32 bits (int)
                
            } else {    // Word 6.0/95 compatible format
                
                _fcClx = LittleEndian.getInt(_mainStream, 0x0160);         // 32 bits (int)
            }
            
            // Which table stream is valid? If (_fWhichTblStm), valid table stream
            // is named "1Table". Otherwise, valid table stream is named "0Table".
            // Normally, a file will have only one table stream, but under unusual
            // circumstances (e.g. crash during file save), a file may have table
            // streams with both names. In that case, this flag must be used to
            // decide which table stream is valid.
            // This flag is NOT used in Word 6.0/95 compatible format, and the piece
            // table is located within the main stream.
            if (isWord97Format()) {    // Word 8.0/97 compatible format
                
                _fWhichTblStm = (flags & 0x200) != 0;
                
            } else {    // Word 6.0/95 compatible format don't use this flag
                
                _fWhichTblStm = false;
            }
        }
    }
    
    /**
     * Parse document content.
     *
     *     @param    in    input stream containing document
     */
    public void parse(InputStream in) {
        
        try {
            
            // Reset handler state becaused it could be used previously.
            reset();
            
            // Create OLE 2.0 file system using the DOC file input stream.
            filesystem = new POIFSFileSystem(in);
            
            // Read document properties.
            readProperties();
            
            // Get important data from the document header block.
            readFIB();
            
            // Extract text from the document.
            if (_mainStream != null) {
                
                if (_fComplex) {
                    
                    //System.out.print("Complex document.\n");
                    processComplexDocument();
                    
                } else {
                    
                    //System.out.print("Non-complex document.\n");
                    processNonComplexDocument();
                }
            }
            removeControlCharacters();
            
        } catch (Exception e) {
            
            e.printStackTrace();
        }
    }
    
    /**
     * Reset handler state.
     */
    private void reset() {

        // Content data        
        title = null;
        description = null;
        keywords = null;
        categories = null;
        published = -1;
        author = null;
        contents.setLength(0);
    }
    
    /**
     * Test.
     */
    public static void main(String[] args) {
        
        if (args.length != 2) {
            
            System.out.println("Arguments:   in.doc   out.txt");
            
        } else {
            
            try {
                
                // Process the input file.
                DOCHandler dh = new DOCHandler();
                dh.parse(new java.io.FileInputStream(args[0]));
                
                // Write properties to stdout.
                System.out.println("Title: " + dh.getTitle());
                System.out.println("Description: " + dh.getDescription());
                System.out.println("Keywords: " + dh.getKeywords());
                System.out.println("Categories: " + dh.getCategories());
                System.out.println("Published: " + dh.getPublished());
                System.out.println("Author: " + dh.getAuthor());
                
                // Write contetns to output file.
                java.io.BufferedWriter bw = new java.io.BufferedWriter(
                        new java.io.FileWriter(args[1]));
                //bw.write("");    // unicode text file marker
                bw.write(dh.getContents(), 0, dh.getContents().length());
                bw.flush();
                bw.close();
                
            } catch(Exception e) {
                
                e.printStackTrace();
            }
        }
    }
}
